#include <YoggieEngine.h>

#include "Renderer.h"
#include "../Scene/Components.h"
#include "../Graphics/Memory/Buffer.h"
#include "../Graphics/Memory/VertexArray.h"
#include "../Graphics/Primitives/DrawCommand.h"

namespace YoggieEngine {
Camera cam = Camera(glm::vec3(12.0f, 0.0f, 0.0f), glm::vec3(45.0f, 0.0f, 0.0f), 90.0f);

Renderer::Renderer(RendererConfig& config)
	: m_framebuffer(Framebuffer(config.ScreenWidth, config.ScreenHeight))
{	
	
}

Renderer::~Renderer(){}


Camera& Renderer::getCamera() {
	return cam;
}

void Renderer::Submit( Render3DComponent& renderComponent, TransformComponent& transform) {
	
	if (renderComponent.VAO == 0 || renderComponent.IBO == 0) 
	{
		if (renderComponent.VAO != 0)
			glDeleteVertexArrays(1, &(renderComponent.VAO));
		if (renderComponent.IBO != 0)
			glDeleteBuffers(1, &(renderComponent.IBO));

		VertexArray va = VertexArray();
		Buffer vertexBuffer = Buffer();
		Buffer elementBuffer = Buffer();

		va.Create();
		va.Bind();

		vertexBuffer.createBuffer();
		vertexBuffer.Bind(false);
		vertexBuffer.setBufferData((void*)&renderComponent.mesh.vertices[0], renderComponent.mesh.vertices.size() * sizeof(Vertex), false);

		elementBuffer.createBuffer();
		elementBuffer.Bind(true);
		elementBuffer.setBufferData((void*)&renderComponent.mesh.elements[0], renderComponent.mesh.elements.size() * sizeof(unsigned int), true);

		va.AttachAttribute(0, 3, sizeof(Vertex));

		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)0);
		glEnableVertexAttribArray(0);
		glEnableVertexAttribArray(1);

		va.Unbind();
		vertexBuffer.Unbind(false);
		elementBuffer.Unbind(true);

		renderComponent.VAO = va.getID();
		renderComponent.IBO = elementBuffer.getBufferID();
	}

	DrawCommand dc = { renderComponent.VAO, renderComponent.mesh.elements.size(), renderComponent.IBO, transform, renderComponent.shader };
	commands.push_back(dc);
	
}

void Renderer::Render() 
{
	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer.GetId());

	if (m_depthTest) {
		glEnable(GL_DEPTH_TEST);
	}

	glClearColor(m_clearColor.r, m_clearColor.g, m_clearColor.b, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	for (const DrawCommand& command : commands)
	{
		glBindVertexArray(command.VAO_identifier);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, command.IBO_identifier);

		command.shader.Use();

		glm::mat4 rotation = glm::rotate(glm::mat4(1.0f), command.transform.Rotation.x, glm::vec3(1.0f, 0.0f, 0.0f));
		rotation *= glm::rotate(glm::mat4(1.0f), command.transform.Rotation.y, glm::vec3(0.0f, 1.0f, 0.0f));
		rotation *= glm::rotate(glm::mat4(1.0f), command.transform.Rotation.z, glm::vec3(0.0f, 0.0f, 1.0f));


		glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), command.transform.Position) * glm::scale(glm::mat4(1.0f), command.transform.Scale) * rotation;


		command.shader.setUniformVec3("Color", glm::vec3(0.3f, 0.3f, 0.3f));
		command.shader.setUniformMat4("M", modelMatrix);
		command.shader.setUniformMat4("V", cam.ViewMatrix);
		command.shader.setUniformMat4("P", cam.ProjectionMatrix);

		glDrawElements(GL_TRIANGLES, static_cast<unsigned int>(command.num_elements),
			GL_UNSIGNED_INT, NULL);

		glBindVertexArray(0);

	}


	/*
	* Lighting pass 
	* 	auto lights = scene.getReg().view<LightComponent>();
		lights.each([&](auto entity, LightComponent& light) {
		  	renderComponent.shader.setUniformVec3("lighting.color", light.Color);
			renderComponent.shader.setUniformFloat("lighting.strength", light.Strength);
		});
	* 
	*/
	
	commands.clear();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}



void Renderer::setCurrentFrameBuffer(const Framebuffer& fb)
{
	m_framebuffer = fb;
}

void Renderer::setClearColor(const glm::vec3& ClearColor)
{
	m_clearColor = ClearColor;
}


}